
#include "regmap.h"

#define DRAM_START	((unsigned *)0x80000000)
#define DRAM_END	((unsigned *)0x84000000)
#define STEP_BYTES	(256*1024)
#define STEP_WORDS	(STEP_BYTES / 4)
#define PATTERN		(0x12345678)

extern void flush_v3(void);
extern void flush_v4(void);

static unsigned char get_char(void)
{
	while(!(readl(UART1_BASE_ADDR + USR1) & 0x200));
	return (readl(UART1_BASE_ADDR + URXD) & 0xff);
}

static void put_char(unsigned char c)
{
	while(!((readl(UART1_BASE_ADDR + USR2) & 0x4000)));
	writel(UART1_BASE_ADDR + UTXD, c);
}

static unsigned get_word(void)
{
	unsigned w;
	
	w = get_char();
	w |= get_char() << 8;
	w |= get_char() << 16;
	w |= get_char() << 24;
	return w;
}

static void put_word(unsigned w)
{
	put_char(w);
	put_char(w >> 8);
	put_char(w >> 16);
	put_char(w >> 24);
}

/*
 * This is a destructive test for DRAM, which is safe because this program
 * runs entirely from internal SRAM.  First we detect the width of the DRAM,
 * which is reported by writing a single byte value of 16 or 32.  This is
 * followed by the compiled-in step size for scanning, which is 256kB.
 * 
 * We start at the end of the DRAM space and work backwards in 256kB steps,
 * writing the address of each block into its first word.  Then we scan
 * forward looking for blocks which contain their own address.  These are
 * the unique blocks of physical memory, which we report with word addresses.
 * The list is terminated with a zero address.
 */
static void detect_dram(void)
{
	volatile unsigned *p;

	put_char(32);
	put_word(STEP_BYTES);

	p = DRAM_END;
	while (p > DRAM_START) {
		p -= STEP_WORDS;
		*p = (unsigned)p;
	}

	/* p == DRAM_START */
	while (p < DRAM_END) {
		if (*p == (unsigned)p) {
			put_word((unsigned)p);
		}
		p+= STEP_WORDS;
	}

	put_word(0);
}


/*
 * Received:	start address
 *		length
 *
 * Transmitted:	data bytes
 *		checksum byte (sum of transmitted data)
 */
static void read_block(void)
{
	unsigned char checksum = 0;
	unsigned char *p = (unsigned char*) get_word();
	unsigned length = get_word();

	while (length--) {
		put_char(*p);
		checksum += *p++;
	}
	put_char(checksum);
}

/*
 * Received:	start address
 *		length
 *		data bytes...
 *
 * Transmitted:	checksum byte (sum of received data)
 */
static void write_block(void)
{
	unsigned char checksum = 0;
	unsigned char *p = (unsigned char*) get_word();
	unsigned length = get_word();

	while (length--) {
		*p = get_char();
		checksum += *p++;
	}
	put_char(checksum);
}

static void 
uart_setup(void)
{
	unsigned int val;
	val = readl(CCM_BASE_ADDR + CCMR);
	writel(CCM_BASE_ADDR + CCMR, val & ~0x1000000);

	writel(UART1_BASE_ADDR + UCR3, 0x0704);
	writel(UART1_BASE_ADDR + UFCR, 0x0a01);
	writel(UART1_BASE_ADDR + ONEMS, 0x81ea);

	writel(UART1_BASE_ADDR + UBIR, 0x0fa2);
	writel(UART1_BASE_ADDR + UBMR, 0xfe80);

	writel(UART1_BASE_ADDR + UCR1, 0x1);
	writel(UART1_BASE_ADDR + UCR2, 0x4027);
	writel(UART1_BASE_ADDR + USR2, 0xffff);
	writel(UART1_BASE_ADDR + USR1, 0xffff);
}

int
main(void)
{
	void (*code)(int r0, int r1, int r2, int r3);
	int r0, r1, r2, r3;
	volatile unsigned *w;
	volatile unsigned char *b;

	/* LED RED ON*/
	*(volatile unsigned int *)0x53fd0000 = 0x00020000;

	uart_setup();

	while (1) {
		switch (get_char()) {
		case '3':	/* Flush v3 MMU */
			flush_v3();
			break;
			
		case '4':	/* Flush v4 MMU */
			flush_v4();
			break;

		case 'C':	/* Clock setting */
			break;

		case 'B':	/* Change UART baudrate 115200bps */
			break;

		case 'a':	/* Ack please */
			put_char('+');	/* we're fine, thanks */
			break;

		case 'c':	/* Call (address, r0, r1, r2, r3) */
			code = (void *)get_word();
			r0 = get_word();
			r1 = get_word();
			r2 = get_word();
			r3 = get_word();

			code(r0, r1, r2, r3);
			break;

		case 'd':	/* Detect DRAM */
			detect_dram();
			break;

		case 'g':	/* Get single byte (address) */
			put_char(*(unsigned char *)get_word());
			break;
		
		case 's':	/* Set single byte (address, data) */
			b = (unsigned char *)get_word();
			*b = get_char();
			break;
		
		case 'r':	/* Read single word (address) */
			put_word(*(unsigned *)get_word());
			break;
		
		case 'w':	/* Write single word (address, data) */
			w = (unsigned *)get_word();
			*w = get_word();
			break;

		case 'R':	/* Read block */
			read_block();
			break;
		
		case 'W':	/* Write block */
			write_block();
			break;

		default:
			put_char('?');
			break;
		}
	}
}
